---------------------------------------------------------------------------
--Arcball Module
---------------------------------------------------------------------------
local Public, Private = {}, {}

Arcball = Public

Private.eventtable = { index = {} }

function Private.eventtable.index.set(a, xc, yc, r, q)
	a.xc = xc or a.xc or 0.5
	a.yc = yc or a.yc or 0.5
	a.r  = r  or a.r or 1
	a.q  = q  or a.q or Quat.fromrotation(0, 0, 1, 0)
	a.qn = Quat.fromrotation(0, 0, 1, 0)
end

function Private.eventtable.index.reset(a)
	a.q  = Quat.fromrotation(0, 0, 1, 0)
	a.qn = Quat.fromrotation(0, 0, 1, 0)
end

function Private.eventtable.index.down(a, x, y)
	a.ei = %Private.getv(a, x, y)
end

function Private.eventtable.index.motion(a, x, y)
	a.ef = %Private.getv(a, x, y)
	local vx, vy, vz = %Private.vcrossp(
        a.ei.x, a.ei.y, a.ei.z, 
        a.ef.x, a.ef.y, a.ef.z
    )
	local vw = %Private.vdotp(
        a.ei.x, a.ei.y, a.ei.z, 
        a.ef.x, a.ef.y, a.ef.z
    )
	a.qn = Quat.create(vx, vy, vz, vw)
end

function Private.eventtable.index.up(a, x, y)
	a.ei = nil
	a.q = Quat.mul(a.qn, a.q)
    a.qn = Quat.fromrotation(0, 0, 1, 0)
end

function Private.eventtable.index.torotation(a)
	return Quat.torotation(Quat.mul(a.qn, a.q))
end

function Public.create(xc, yc, r, q)
	local a = {}
    eventtable(a, Private.eventtable)
    a:set(xc, yc, r, q)
	return a
end

function Private.getv(a, x, y)
	local x = (x - a.xc) / a.r
	local y = (y - a.yc) / a.r
	local n = x*x + y*y
	local e = {}
	if n >= 1 then
		n = sqrt(n)
		e.x = x / n
		e.y = y / n
		e.z = 0
	else
		e.x = x
		e.y = y
		e.z = sqrt(1 - n)
	end
	return e
end	

function Private.vcrossp(x, y, z, a, b, c)
	return y*c - z*b, z*a - x*c, x*b - y*a
end

function Private.vdotp(x, y, z, a, b, c)
	return x*a + y*b + z*c
end
