---------------------------------------------------------------------------
--Arena Module
---------------------------------------------------------------------------
local Private, Public, Local = {}, {}, {}
Arena = Public

function Private.create2DList(steps)
    local list = gl_genlists(1)
    local steps = steps or 25
    local c
    gl_newlist(list, GL_COMPILE)
    gl_enable(GL_COLOR_MATERIAL)
    gl_disable(GL_LIGHTING)
    gl_begin(GL_QUADS)
    for i = 1, steps do
        for j = 1, steps do
            gl_normal(0, 0, 1)
            c = Public.pressure(-1 + 2*(j-1)/steps, -1 + 2*(i-1)/steps)
            gl_color(c, c, c)
            gl_vertex(-1 + 2*(j-1)/steps, -1 + 2*(i-1)/steps, -0.1)
            c = Public.pressure(-1 + 2*j/steps, -1 + 2*(i-1)/steps)
            gl_color(c, c, c)
            gl_vertex(-1 + 2*j/steps, -1 + 2*(i-1)/steps, -0.1)
            c = Public.pressure(-1 + 2*j/steps, -1 + 2*i/steps)
            gl_color(c, c, c)
            gl_vertex(-1 + 2*j/steps, -1 + 2*i/steps, -0.1)
            c = Public.pressure(-1 + 2*(j-1)/steps, -1 + 2*i/steps)
            gl_color(c, c, c)
            gl_vertex(-1 + 2*(j-1)/steps, -1 + 2*i/steps, -0.1)
        end
    end
    gl_end()
    gl_disable(GL_COLOR_MATERIAL)
    gl_enable(GL_LIGHTING)
    gl_endlist()
    return list
end

function Private.create3DList(img, steps, lines)
    local model
    local ret, err = openfile(img, "r")
    if err then return nil, err end
    local texture = gl_rgbtex(rgb_load(img))
    local list = gl_genlists(1)
    local steps = steps or 25
    if Local.model then model = Scene.load(Local.model) end
    gl_newlist(list, GL_COMPILE)
    gl_material(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE, 1, 1, 1, 1.0)
    gl_material(GL_FRONT, GL_SPECULAR, 1, 1, 1, 1.0)
    gl_material(GL_FRONT, GL_SHININESS, 10.0)
    if model then 
        gl_pushmatrix()
            gl_scale(model.factor, model.factor, 2/model.depth)
            gl_translate(-model.xc, -model.yc, -model.zc)
            model:draw()
        gl_popmatrix()
    else
    gl_enable(GL_TEXTURE_2D)
    gl_texenv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE)
    gl_texenv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, 1, 1, 1, 1)
    gl_bindtex(GL_TEXTURE_2D, texture)
    gl_begin(GL_QUADS)
    for i = 1, steps do
        for j = 1, steps do
            gl_normal(0, 0, 1)
            gl_texcoord((j-1)/steps, (i-1)/steps)
            gl_vertex(-1 + 2*(j-1)/steps, -1 + 2*(i-1)/steps, 1*(-1 + 2*
                Public.height(-1 + 2*(j-1)/steps, -1 + 2*(i-1)/steps)))
            gl_texcoord(j/steps, (i-1)/steps)
            gl_vertex(-1 + 2*j/steps, -1 + 2*(i-1)/steps, 1*(-1 + 2*
                Public.height(-1 + 2*j/steps, -1 + 2*(i-1)/steps)))
            gl_texcoord(j/steps, i/steps)
            gl_vertex(-1 + 2*j/steps, -1 + 2*i/steps, 1*(-1 + 2*
                Public.height(-1 + 2*j/steps, -1 + 2*i/steps)))
            gl_texcoord((j-1)/steps, i/steps)
            gl_vertex(-1 + 2*(j-1)/steps, -1 + 2*i/steps, 1*(-1 + 2*
                Public.height(-1 + 2*(j-1)/steps, -1 + 2*i/steps)))
        end
    end
    gl_end()
    end
    gl_begin(GL_QUADS)
        gl_normal(0, 0, -1)
        gl_texcoord(0, 0)
        gl_vertex(-1, -1, -1)
        gl_texcoord(1, 0)
        gl_vertex(1, -1, -1)
        gl_texcoord(1, 1)
        gl_vertex(1, 1, -1)
        gl_texcoord(0, 1)
        gl_vertex(-1, 1, -1)
    gl_end()
    gl_disable(GL_TEXTURE_2D)
    gl_material(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, 0.5, 0.5, 0.5, 1)
    gl_begin(GL_QUADS)
    for i = 1, steps do
        for j = 1, steps do
            gl_normal(0, 1, 0)
            gl_vertex(-1 + 2*(j-1)/steps, 1, -1)
            gl_vertex(-1 + 2*j/steps, 1, -1)
            gl_vertex(-1 + 2*j/steps, 1, 1*(-1 + 2*
                    Public.height(-1 + 2*(j)/steps, 1)))
            gl_vertex(-1 + 2*(j-1)/steps, 1, 1*(-1 + 2*
                    Public.height(-1 + 2*(j-1)/steps, 1)))
            gl_normal(0, -1, 0)
            gl_vertex(-1 + 2*(j-1)/steps, -1, -1)
            gl_vertex(-1 + 2*j/steps, -1, -1)
            gl_vertex(-1 + 2*j/steps, -1, 1*(-1 + 2*
                    Public.height(-1 + 2*(j)/steps, -1)))
            gl_vertex(-1 + 2*(j-1)/steps, -1, 1*(-1 + 2*
                    Public.height(-1 + 2*(j-1)/steps, -1)))
            gl_normal(1, 0, 0)
            gl_vertex(1, -1 + 2*(j-1)/steps, -1)
            gl_vertex(1, -1 + 2*j/steps, -1)
            gl_vertex(1, -1 + 2*j/steps, 1*(-1 + 2*
                    Public.height(1, -1 + 2*(j)/steps)))
            gl_vertex(1, -1 + 2*(j-1)/steps, 1*(-1 + 2*
                    Public.height(1, -1 + 2*(j-1)/steps)))
            gl_normal(-1, 0, 0)
            gl_vertex(-1, -1 + 2*(j-1)/steps, -1)
            gl_vertex(-1, -1 + 2*j/steps, -1)
            gl_vertex(-1, -1 + 2*j/steps, 1*(-1 + 2*
                    Public.height(-1, -1 + 2*(j)/steps)))
            gl_vertex(-1, -1 + 2*(j-1)/steps, 1*(-1 + 2*
                    Public.height(-1, -1 + 2*(j-1)/steps)))
        end
    end
    gl_end()
    if lines then
    gl_material(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, 0, 0, 0, 1)
    for i = 1, steps do
        for j = 1, steps do
            gl_begin(GL_LINES)
                gl_vertex(-1 + 2*(j-1)/steps, -1 + 2*(i-1)/steps, -1 + 2*
                    Public.height(-1 + 2*(j-1)/steps, -1 + 2*(i-1)/steps))
                gl_vertex(-1 + 2*j/steps, -1 + 2*(i-1)/steps, -1 + 2*
                    Public.height(-1 + 2*j/steps, -1 + 2*(i-1)/steps))
                gl_vertex(-1 + 2*j/steps, -1 + 2*(i-1)/steps, -1 + 2*
                    Public.height(-1 + 2*j/steps, -1 + 2*(i-1)/steps))
                gl_vertex(-1 + 2*j/steps, -1 + 2*i/steps, -1 + 2*
                    Public.height(-1 + 2*j/steps, -1 + 2*i/steps))
                gl_vertex(-1 + 2*j/steps, -1 + 2*i/steps, -1 + 2*
                    Public.height(-1 + 2*j/steps, -1 + 2*i/steps))
                gl_vertex(-1 + 2*(j-1)/steps, -1 + 2*i/steps, -1 + 2*
                    Public.height(-1 + 2*(j-1)/steps, -1 + 2*i/steps))
                gl_vertex(-1 + 2*(j-1)/steps, -1 + 2*i/steps, -1 + 2*
                    Public.height(-1 + 2*(j-1)/steps, -1 + 2*i/steps))
                gl_vertex(-1 + 2*(j-1)/steps, -1 + 2*(i-1)/steps, -1 + 2*
                    Public.height(-1 + 2*(j-1)/steps, -1 + 2*(i-1)/steps))
            gl_end()
        end
    end
    end
    gl_endlist()
    return list
end

function Public.deleteList()
    gl_deletelists(Local.list2D, 1)
    gl_deletelists(Local.list3D, 1)
end
    
function Public.draw2D()
    gl_calllist(Local.list2D)
end

function Public.draw3D()
    gl_calllist(Local.list3D)
end

function Public.load(file)
    local description, err = dofile(file)
    if err then return nil, err end
    Local.hmap = mono_createfunc(mono_load(description.hmap))
    Local.pmap = mono_createfunc(mono_load(description.pmap))
    Local.model = description.model
    Public.deleteList()
    Local.list2D = Private.create2DList(15)
    Local.list3D = Private.create3DList(description.face)
    return 1
end

--returns the height in position (x,y)
--(converts [-1, 1] to [0,1])
function Public.height(x, y)
    return Local.hmap((x+1)/2, (y+1)/2)
end

--returns the pressure in position (x,y)
--(converts [-1, 1] to [0,1])
function Public.pressure(x, y)
    return Local.pmap((x+1)/2, (y+1)/2)
end
