---------------------------------------------------------------------------
--Quaternion Module
---------------------------------------------------------------------------
--inits "qt" table
local Public = {}
Quat = Public

--Returns the product of the quaternions qL * qR.
function Public.mul(qL, qR)
	local qq = {}
	qq.w = qL.w*qR.w - qL.x*qR.x - qL.y*qR.y - qL.z*qR.z
	qq.x = qL.w*qR.x + qL.x*qR.w + qL.y*qR.z - qL.z*qR.y
	qq.y = qL.w*qR.y + qL.y*qR.w + qL.z*qR.x - qL.x*qR.z
	qq.z = qL.w*qR.z + qL.z*qR.w + qL.x*qR.y - qL.y*qR.x
	return qq
end

--Returns the product of a quaterinon by a scalar.
function Public.smul(q, a)
	local qq = {}
	qq.w = q.w*a
	qq.x = q.x*a
	qq.y = q.y*a
	qq.z = q.z*a
	return qq
end

--Returns the sum of the quaternions qL + qR.
function Public.add(qL, qR)
	local qq = {}
	qq.w = qL.w + qR.w
	qq.x = qL.x + qR.x
	qq.y = qL.y + qR.y
	qq.z = qL.z + qR.z
	return qq
end

--Returns the norm of a quaternion
function Public.norm(q)
	local x = q.x
	local y = q.y
	local z = q.z
	local w = q.w
	return x*x + y*y + z*z + w*w
end

-- Creates a quaternion
function Public.create(x, y, z, w)
	return { x = x, y = y, z = z, w = w }
end 

--Normalizes a quaternion
function Public.normalize(q)
	local s = sqrt(Public.norm(q))
	return Public.create(q.x / s, q.y / s, q.z / s, q.w / s)
end

--Returns a normalized quaternion from a rotation
function Public.fromrotation(a, x, y, z)
	local n = sqrt(x*x + y*y + z*z)
	if abs(n) < 0.000001 then n = 1 end
	local s = sin(a/2)
	return Public.create(x*s/n, y*s/n, z*s/n, cos(a/2))
end

--Prints the values of a quaternion
function Public.print(q)
	write("{x = ", tonumber(q.x), ", y = ", tonumber(q.y), ", z = ", tonumber(q.z), ", w = ", tonumber(q.w), "}\n")
end

--Returns the values for a rotation(angle, <vector>) from a quaternion
function Public.torotation(q)
	q = Public.normalize(q)
	return 2*acos(q.w), q.x, q.y, q.z
end
