------------------------------------------------------------------------------
-- Simulator Module
------------------------------------------------------------------------------
local Private, Public, Local = {}, {}, {}
Simulator = Public
Local.arena_size = 512
Local.robot_size = 32
Local.probe_size = 64
Local.zscale = 0.2
Public.timeLimit = 180

--- Visibility issues
local Arena = Arena
local Probe = Probe
local GUI   = GUI
---

Local.TeamList = {}
Local.RobotList = {}
Local.ProbeList = {}
Public.DeadBots = {}
Local.privRobotTable = {}
Local.DelayTable = {
    forward = 1,
    back = 1,
    left = 1,
    right = 1,
    position = 1,
    speed = 1,
    deltas = 0,
    pressure = 1,
    probe = 1500,
    print = 0,
}

------------------------------------------------------------------------------
-- Auxiliary functions:
------------------------------------------------------------------------------
function Private.checkScript(script)
    local globals = globals
    local loadfile = loadfile
    local err = ""
    local ng = {
        _ERRORMESSAGE = _ERRORMESSAGE,
        _ALERT = function(s)
            err = s 
        end 
    }
    local g = globals(ng)
    local f, ferr = loadfile(script)
    globals(g)
    if not f then return nil, ferr .. "\n" .. err
    else return 1 end
end

function Private.existFileQ(file)
    local ret, err = openfile(file, "r")
    if not ret and err == "Arquivo no existente" then return nil
    else closefile(ret) return 1 end
end

function Private.tick_wait(delay)
    Local.tick_mutex:lock()
    local releaseTick = Local.tick + Local.DelayTable[delay]
    while (Local.tick < releaseTick) do
        Local.tick_cond:wait(Local.tick_mutex)
    end
    Local.tick_mutex:unlock()
end

function Private.insideQ(x, y, x0, y0, x1, y1)
    if x < x0 then return nil end
    if y < y0 then return nil end
    if x > x1 then return nil end
    if y > y1 then return nil end
    return 1
end

function Private.isEmptyQ(list, x, y, size)
    return Private.countHit(list, x, y, size) < 1
end

function Private.canMoveQ(list, x, y, size)
    return Private.countHit(list, x, y, size) < 2
end

function Private.countHit(list, x, y, size)
    local hit = 0
    local insideQ = Private.insideQ
    local size2 = 2 * size
    local x0, x1, y0, y1
    x0 = x - size2
    x1 = x + size2
    y0 = y - size2
    y1 = y + size2
    for i = 1, getn(list) do
        local o = list[i]
        if insideQ(o.x, o.y, x0, y0, x1, y1) then
            hit = hit + 1
        end
    end
    return hit
end

local slot = 0
function Private.freeRandomSlot()
    slot = slot + 1
    local arena = Local.bound * Local.arena_size
    local size = Local.w_robot_size
    local list = Local.RobotList
    local step = Local.w_step_size
    local x, y 
    repeat 
       x = random(-arena, arena) * step 
       y = random(-arena, arena) * step 
    until Private.isEmptyQ(list, x, y, size)
    return x, y
end

function Private.getSlot()
    if not slot then  print("o qu? mais de 4 robs???") exit() end
    slot = slot + 1
    local arena = Local.bound * Local.arena_size
    local step = Local.w_step_size
    if slot == 1 then
        x = -arena
        y = arena
    elseif slot == 2 then
        x = arena
        y = arena
    elseif slot == 3 then
        x = -arena
        y = -arena
    elseif slot == 4 then
        x = arena
        y = -arena
    end
    return x*step, y*step
end

Private.getSlotFunction = Private.getSlot

function Private.getSpeed()
    local step = Local.w_step_size
    if slot == 1 then
        vx = 0
        vy = -step
    elseif slot == 2 then
        vx = -step
        vy = 0
    elseif slot == 3 then
        vx = step
        vy = 0
    elseif slot == 4 then
        vx = 0
        vy = step
        slot = nil
    end
    return vx, vy
end

function Private.randomSpeed()
    local vx, vy = 0, Local.w_step_size
    for i = 1, random(0, 3) do
        local vt = vx
        vx = vy
        vy = -vt 
    end
    return vx, vy
end

function Private.toArena(x, y)
    local ax = x * Local.arena_size
    local ay = y * Local.arena_size
    return ax, ay
end

function Private.launchRobot(userRobot, script)
    local globals = globals
    local newthread = newthread
    local dofile = dofile
    local ng = { 
        robo = userRobot,
        robot = userRobot,
        _ERRORMESSAGE = _ERRORMESSAGE,
        _ALERT = function(s)
            robo:kill(s)
        end,
        write = write,
        _STDERR = _STDERR,
        floor = floor,
        getn = getn,
        random = random,
        abs = abs,
        tinsert = tinsert,
        tremove = tremove,
        max = max,
	min = min,
        mod = mod,
        sqrt = sqrt,
        pow = pow,
	TAMANHO = Local.arena_size*2,
	SIZE = Local.arena_size*2,
        TEMPO = Simulator.timeLimit,
        TIME = Simulator.timeLimit
    }
    local g = globals(ng)
    newthread(dofile, {script})
    globals(g)
end

function Private.interangle(a, b, s)
    local d1 = abs(a - b)
    local d2 = 360 - d1
    if d1 > d2 then
        if a > b then b = a + d2
        else b = a - d2 end
    end
    return a*s + b*(1-s)
end

function Private.getangle(x, y, angle)
    local a, b, r, diff
    r = Local.w_robot_size
    a = r*cos(angle)
    b = r*sin(angle)
    diff = Arena.height(x+a, y+b) - Arena.height(x-a, y-b)
    return atan2(diff*2*Local.zscale, r*2)
end

local atan2 = atan2
function Private.getPAngle(x, y)
    local hps = Local.w_probe_size
    local ps = 2*hps
    local dx = Arena.height(x, y+hps) - Arena.height(x, y-hps)
    dx = 2*dx*Local.zscale
    local dy = Arena.height(x+hps, y) - Arena.height(x-hps, y)
    dy = 2*dy*Local.zscale
    local x = atan2(dx, ps)
    local y = -atan2(dy, ps)
    return x, y
end

function Private.addScores()
    local probes = Local.ProbeList
    for i = 1, getn(probes) do
        local probe = probes[i]
        probe.team.score = probe.team.score + probe.pressure^4
    end
end

function Public.showScores()
    local teams = Local.TeamList
    local winner = { name="Empate", score=0 }
    print()
    for i = 1, getn(teams) do
        local team = teams[i]
	if team.score>winner.score then
	  winner=team
	end
        print(team.name, "produo: " .. team.score)
    end
    print("VENCEDOR: " .. winner.name)
end

------------------------------------------------------------------------------
--External Functions: 
------------------------------------------------------------------------------

---Adds a Team to the team table
function Public.addTeam(name, color, nrobots, script)
    ---=Checks if everything is OK.
    local convert = function(s)
        local r,g,b,_
        _, _, r, g, b = strfind(s, "(%d*) (%d*) (%d*)")
        if r and g and b then return {r = r/255, g = g/255, b = b/255}
        else return {r = 0, g = 0, b = 0} end
    end
    for i = 1, getn(Local.TeamList) do
        team = Local.TeamList[i]
        if team.name == name then
            return nil, "Nome j usado"
        end 
    end
    nrobots = tonumber(nrobots)
    if not nrobots or nrobots < 1 or nrobots > 16 then 
        return nil, "Nmero de robs deve ser entre 1 e 16."
    end
    local ret, err = Private.checkScript(script)
    if not ret then return nil, err end
    GUI.teamOK = 1
    ---=Inserts the fields in the Team Table.
    tinsert(Local.TeamList, {
        name    = name,
        color   = convert(color),
        nrobots = nrobots,
        script  = script,
        score   = 0
    }) 
    return 1
end

function Public.init()
    Local.w_robot_size = Local.robot_size / Local.arena_size
    Local.w_step_size = 1 / Local.arena_size
    Local.w_probe_size = Local.probe_size * Local.w_step_size
    Local.bound = 1 - Local.w_robot_size
    Local.tick_mutex = newmutex()
    Local.tick_cond = newcond()
    Local.mutex = newmutex()
    Local.robotAnim = Animation.load("robot.anm")
    Local.probeAnim = Animation.load("probe.anm")
    --ProbeAnim.init()
end

function Public.clean()
    Local.privRobotTable = {}
    Local.RobotList = {} 
    Local.TeamList = {}
end

function Public.toggleSlotFunction()
    if Private.getSlotFunction == Private.freeRandomSlot then
        Private.getSlotFunction = Private.getSlot
    else
        Private.getSlotFunction = Private.freeRandomSlot
    end
end

function Public.restart()
    Public.tick(0)
    Local.privRobotTable = {}
    Local.RobotList = {} 

    Local.mutex:lock()
    for i = 1, getn(Local.TeamList) do
        team = Local.TeamList[i]
        for j = 1, team.nrobots do
            local userRobot = userRobot.create(j)
            local x, y = Private.getSlotFunction()
	   print(x) print(y)
            local vx, vy = Private.getSpeed()
            local newPrivRobot = privRobot.create(x, y, vx, vy, team, 
                Local.robotAnim.moving)
            Local.privRobotTable[userRobot] = newPrivRobot
            tinsert(Local.RobotList, newPrivRobot)
            Private.launchRobot(userRobot, team.script)
        end
    end
    Local.mutex:unlock()
    RobotList = Local.RobotList
end

---Resets Team table
function Public.reset()
    Local.TeamList = {}
    Local.ProbeList = {}
    Local.RobotList = {}
    Local.privRobotTable = {}
    Arena.deleteList()
end

function Public.tick(value)
    Local.tick_mutex:lock()
    value = value or Local.tick + 1
    Private.addScores()
    Local.tick = value
    Local.tick_cond:broadcast()
    Local.tick_mutex:unlock()
    -- adjust positions and orientations
    Local.mutex:lock()
    local robotList = Local.RobotList
    local probeList = Local.ProbeList
    local rsize = Local.w_robot_size
    local psize = Local.w_probe_size
    for i = 1, getn(robotList) do
        local robot = robotList[i]
        local steps = 1/(4*Local.robot_size)
        local x = robot.x * steps + robot.x0 * (1 - steps)
        local y = robot.y * steps + robot.y0 * (1 - steps)
        local dx = x - robot.x0
        local dy = y - robot.y0
        local angle = Private.interangle(atan2(dy, dx), robot.angle0, steps)
        robot.x0 = x
        robot.y0 = y
        robot.angle0 = angle
    end
    --advances 1 frame:
    for i = 1, getn(probeList) do
        local probe = probeList[i]
        probe:nextFrame()
    end
    Local.mutex:unlock()
end

function Public.drawSchematic()
    Local.mutex:lock()
    local robotList = Local.RobotList
    local probeList = Local.ProbeList
    local rsize = Local.w_robot_size
    local psize = Local.w_probe_size
    --Draws the Arena
    Arena.draw2D()
    --Draws the Robots
    gl_enable(GL_COLOR_MATERIAL)
    gl_disable(GL_LIGHTING)

    for i = 1, getn(robotList) do
        local robot = robotList[i]
        local x = robot.x0
        local y = robot.y0
        local rx = robot.x
        local ry = robot.y
        local color = robot.team.color
        local angle = robot.angle0
        gl_pushmatrix()
            gl_translate(x, y, 0.1)
            gl_scale(rsize, rsize, rsize)
            gl_rotate(angle, 0, 0, 1) 
            gl_color(color.r, color.g, color.b)
            gl_begin(GL_TRIANGLES)
                gl_normal(0, 0, 1)
                gl_vertex(-1, -1, 0)
                gl_vertex(1, 0, 0)
                gl_vertex(-1, 1, 0)
            gl_end()
        gl_popmatrix()

        if GUI.detailed then
            gl_pointsize(5)
            gl_begin(GL_POINTS)
                gl_normal(0, 0, 1)
                gl_vertex(rx, ry, 0)
            gl_end()
            gl_begin(GL_LINE_STRIP)
                gl_normal(0, 0, 1)
                gl_vertex(x, y, 0)
                gl_vertex(rx, ry, 0)
                gl_vertex(rx+robot.vx*30, ry+robot.vy*30, 0)
            gl_end()
        end
    end
    --Draws the Probes
    for i = 1, getn(probeList) do
        local probe = probeList[i]
        local color = probe.team.color
        local x, y = probe.x, probe.y
        gl_pushmatrix()
            gl_translate(x, y, 0)
            gl_color(color.r, color.g, color.b)
            gl_pointsize(5)
            gl_begin(GL_POINTS)
                gl_normal(0, 0, 1)
                gl_vertex(0, 0, 0)
            gl_end()
            gl_begin(GL_LINE_LOOP)
                gl_normal(0, 0, 1)
                gl_vertex(-psize, psize, 0)
                gl_vertex(-psize, -psize, 0)
                gl_vertex(psize, -psize, 0)
                gl_vertex(psize, psize, 0)
            gl_end()
        gl_popmatrix()
    end
    gl_disable(GL_COLOR_MATERIAL)
    gl_enable(GL_LIGHTING)
    --Draws the Scores
    Score.draw(Local.TeamList)
    Local.mutex:unlock()
end

function Public.draw3D()
    Local.mutex:lock()
    local robotList = Local.RobotList
    local probeList = Local.ProbeList
    local rsize = Local.w_robot_size
    local psize = Local.w_probe_size
    --Draws the Arena
    gl_pushmatrix()
        gl_scale(1, 1, Local.zscale)
        Arena.draw3D()
    gl_popmatrix()
    --Draws the Robots
    for i = 1, getn(robotList) do
        local robot = robotList[i]
        local x = robot.x0
        local y = robot.y0
        local color = robot.team.color
        local angle = robot.angle0
        local yangle = Private.getangle(x, y, angle)
        gl_pushmatrix()
            gl_translate(x, y, Local.zscale*(-1+2*(Arena.height(x,y)))+0.1)
            gl_scale(rsize, rsize, rsize)
            gl_rotate(angle, 0, 0, 1) 
            gl_rotate(yangle, 0, -1, 0)
            gl_material(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE,
                color.r, color.g, color.b)
            robot:draw()
        gl_popmatrix()
    end
    --Draws the Probes
    for i = 1, getn(probeList) do
        local probe = probeList[i]
        local color = probe.team.color
        local x, y = probe.x, probe.y
        gl_pushmatrix()
            gl_material(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE,
                color.r, color.g, color.b, 1)
            gl_translate(x, y, Local.zscale*(-1+2*(Arena.height(x,y)))+
            psize/4)
            gl_begin(GL_LINE_LOOP)
                gl_normal(0, 0, 1)
                gl_vertex(-psize, psize, 0)
                gl_vertex(-psize, -psize, 0)
                gl_vertex(psize, -psize, 0)
                gl_vertex(psize, psize, 0)
            gl_end()
            gl_scale(psize/4, psize/4, psize/4)
            probe:draw()
        gl_popmatrix()
    end
    --Draws the Scores
    Score.draw(Local.TeamList)
    Local.mutex:unlock()
end

------------------------------------------------------------------------------
--Robot Functions: 
------------------------------------------------------------------------------

function Public.forward(userRobot)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    local x = privRobot.x + privRobot.vx
    local y = privRobot.y + privRobot.vy
    local ret
    local b = Local.bound
    if Private.insideQ(x, y, -b, -b, b, b) and
    Private.canMoveQ(Local.RobotList, x, y, Local.w_robot_size) then
        ret = 1
        privRobot.x = x
        privRobot.y = y
    end
    privRobot:nextFrame()
    privRobot:setAnimation(Local.robotAnim.moving)
    Local.mutex:unlock()
    Private.tick_wait("forward")
    return ret
end

function Public.back(userRobot)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    local x = privRobot.x - privRobot.vx
    local y = privRobot.y - privRobot.vy
    local ret
    local b = Local.bound
    if Private.insideQ(x, y, -b, -b, b, b) and
    Private.canMoveQ(Local.RobotList, x, y, Local.w_robot_size) then
        ret = 1
        privRobot.x = x
        privRobot.y = y
    end
    privRobot:nextFrame()
    privRobot:setAnimation(Local.robotAnim.moving)
    Local.mutex:unlock()
    Private.tick_wait("back")
    return ret
end

function Public.left(userRobot)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    local vx = -privRobot.vy
    privRobot.vy = privRobot.vx
    privRobot.vx = vx
    Local.mutex:unlock()
    Private.tick_wait("left")
    return 1
end

function Public.right(userRobot)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    local vx = privRobot.vy
    privRobot.vy = -privRobot.vx
    privRobot.vx = vx
    Local.mutex:unlock()
    Private.tick_wait("right")
    return 1
end

function Public.position(userRobot)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    local x, y = Private.toArena(privRobot.x, privRobot.y)
    x = floor(x + 0.5)
    y = floor(y + 0.5)
    Local.mutex:unlock()
    Private.tick_wait("position")
    return x, y
end

function Public.speed(userRobot)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    local vx, vy = Private.toArena(privRobot.vx, privRobot.vy)
    Local.mutex:unlock()
    Private.tick_wait("speed")
    return vx, vy
end

function Public.deltas(userRobot)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    local dt, dn
    local x, y 
    local vx, vy
    local s = 16
    x, y = privRobot.x, privRobot.y
    vx, vy = s*privRobot.vx, s*privRobot.vy
    dt = Arena.pressure(x+vx, y+vy) - Arena.pressure(x-vx, y-vy)
    dn = Arena.pressure(x+vy, y-vx) - Arena.pressure(x-vy, y+vx)
    Local.mutex:unlock()
    Private.tick_wait("deltas")
    return dt, dn
end

function Public.pressure(userRobot)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    local pressure = Arena.pressure(privRobot.x, privRobot.y)
    Local.mutex:unlock()
    Private.tick_wait("pressure")
    return pressure
end

function Public.probe(userRobot)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    local x = privRobot.x
    local y = privRobot.y
    local team = privRobot.team
    local ret
    if Private.isEmptyQ(Local.ProbeList, x, y, Local.w_probe_size) then
        ret = 1
        local newProbe = Probe.create(x, y, team, 
            Arena.pressure(x, y), Local.probeAnim.idle,
            Private.getPAngle(x, y))
        tinsert(Local.ProbeList, newProbe)
    end
    if ret then 
        privRobot:nextFrame() 
        privRobot:setAnimation(Local.robotAnim.drilling) 
    end
    Local.mutex:unlock()
    if ret then Private.tick_wait("probe") end
    return ret
end

function Public.sendmessage(userRobot, tipo, str)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    local this
    for i = 1, getn(Local.RobotList) do
        this = Local.RobotList[i]
        if this.team.name == privRobot.team.name and this ~= privRobot then
            this:putMsg(tipo, str)
        end
    end
    Local.mutex:unlock()
    return 1
end

function Public.getmessage(userRobot)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    Local.mutex:unlock()
    return privRobot:getMsg()
end

function Public.time()
    Local.mutex:lock()
    local ret = GUI.timeElapsed()
    Local.mutex:unlock()
    return ret
end

function Public.print(userRobot, ...)
    Local.mutex:lock()
    local privRobot = Local.privRobotTable[userRobot]
    local s = "[" .. privRobot.team.name .. "(" .. userRobot.id .. ")]: "
    local i = 1
    while arg[i] do
        s = s .. arg[i] .. " "
        i = i + 1
    end
    s = s .. "\n"
    write(_STDERR, s)
    Local.mutex:unlock()
end

function Public.kill(userRobot, err)
    Local.mutex:lock()
        local privRobot = Local.privRobotTable[userRobot]
        privRobot:setAnimation(Local.robotAnim.broken) 
        tinsert(Public.DeadBots, {
            name = userRobot.id,
            team = privRobot.team.name,
            err  = err })
    Local.mutex:unlock()
end

